-- This Source Code Form is subject to the terms of the bCDDL, v. 1.1.
-- If a copy of the bCDDL was not distributed with this
-- file, You can obtain one at http://beamng.com/bCDDL-1.1.txt

local M = {}
M.type = "auxiliary"
M.relevantDevice = "transfercase"

local driveModesTRLT_Range = nil
local driveModesTransfercase = nil
local driveModesDifferentials = nil

local TRLT_shaft = nil
local TRLT_rangeBox = nil
local hasBuiltPie = false

local function updateGFX(dt)
  electrics.values.modeTRLT_RangeBox = TRLT_rangeBox and (TRLT_rangeBox.mode == "low" and 1 or 0) or 0
  electrics.values.modeTRLT_4WD = TRLT_shaft and (TRLT_shaft.mode == "connected" and 1 or 0) or 0
end

local function toggleDiffs()
  if driveModesDifferentials then
    for _, c in ipairs(driveModesDifferentials) do
      c.nextDriveMode()
    end
  else
    --backwards compat code for older vehicles that only use the 4wd controller and no drivemodes
    --input action calls this method if a 4wd controller is found, if not it directly call the powertrain API
    powertrain.toggleDefaultDiffs()
  end
end

local function toggleTRLT_Range()
  if driveModesTRLT_Range then
    for _, c in ipairs(driveModesTRLT_Range) do
      c.nextDriveMode()
    end
  else
    --backwards compat code for older vehicles that only use the 4wd controller and no drivemodes
    if TRLT_rangeBox then
      powertrain.toggleDeviceMode(TRLT_rangeBox.name)
    end
  end
end

local function setTRLT_RangeMode(mode)
  if TRLT_rangeBox then
    powertrain.setDeviceMode(TRLT_rangeBox.name, mode)
  end
end

local function toggleTRLT_4WD()
  if driveModesTransfercase then
    for _, c in ipairs(driveModesTransfercase) do
      c.nextDriveMode()
    end
  else
    --backwards compat code for older vehicles that only use the 4wd controller and no drivemodes
    if TRLT_shaft and not TRLT_shaft.isPhysicallyDisconnected then
      powertrain.toggleDeviceMode(TRLT_shaft.name)
    end
  end
end

local function setTRLT_4WDMode(mode)
  if TRLT_shaft and not TRLT_shaft.isPhysicallyDisconnected then
    powertrain.setDeviceMode(TRLT_shaft.name, mode)
  end
end

local function serialize()
  return {
    modeTRLT_4WD = TRLT_shaft and TRLT_shaft.mode or nil,
    modeTRLT_Range = TRLT_rangeBox and TRLT_rangeBox.mode or nil
  }
end

local function deserialize(data)
  if data then
    if TRLT_shaft and data.modeTRLT_4WD then
      setTRLT_4WDMode(data.modeTRLT_4WD)
    end
    if TRLT_rangeBox and data.modeTRLT_Range then
      setTRLT_RangeMode(data.modeTRLT_Range)
    end
  end
end

local function init(jbeamData)
  TRLT_shaft = powertrain.getDevice(jbeamData.TRLT_shaftName)
  TRLT_rangeBox = powertrain.getDevice(jbeamData.TRLT_rangeBoxName)

  if TRLT_shaft then
    electrics.values.modeTRLT_4WD = TRLT_shaft.mode == "connected" and 1 or 0
  end
  if TRLT_rangeBox then
    electrics.values.modeTRLT_RangeBox = TRLT_rangeBox.mode == "low" and 1 or 0
  end

  if not hasBuiltPie then
    if TRLT_shaft then
      core_quickAccess.addEntry(
        {
          level = "/powertrain/",
          generator = function(entries)
            local wdIcon
            if TRLT_shaft.mode == "disconnected" then
              wdIcon = "radial_disconnected"
            else
              wdIcon = "radial_connected"
            end

            local wdEntry = {
              title = "ui.radialmenu2.powertrain.TRLT_4WD_Mode",
              icon = wdIcon,
              onSelect = function()
                if TRLT_shaft.mode == "disconnected" then
                  controller.getController(M.name).setTRLT_4WDMode("connected")
                else
                  controller.getController(M.name).setTRLT_4WDMode("disconnected")
                end
                return {"reload"}
              end
            }
            if TRLT_shaft.mode == "connected" then
              wdEntry.color = "#ff6600"
            end
            table.insert(entries, wdEntry)
          end
        }
      )
    end
    if TRLT_rangeBox then
      core_quickAccess.addEntry(
        {
          level = "/powertrain/",
          generator = function(entries)
            local rmIcon
            if TRLT_rangeBox.mode == "low" then
              rmIcon = "radial_lowrangebox"
            else
              rmIcon = "radial_highrangebox"
            end

            local rmEntry = {
              title = "ui.radialmenu2.powertrain.rangebox_mode",
              icon = rmIcon,
              onSelect = function()
                if TRLT_rangeBox.mode == "low" then
                  controller.getController(M.name).setTRLT_RangeMode("high")
                else
                  controller.getController(M.name).setTRLT_RangeMode("low")
                end
                return {"reload"}
              end
            }
            rmEntry.color = "#ff6600"
            table.insert(entries, rmEntry)
          end
        }
      )
    end
    hasBuiltPie = true
  end
end

local function initLastStage(jbeamData)
  local driveModesTRLT_RangeNames = jbeamData.driveModesTRLT_RangeNames or {}
  local driveModesTransfercaseNames = jbeamData.driveModesTransfercaseNames or {}
  local driveModesDifferentialNames = jbeamData.driveModesDifferentialNames or {}

  driveModesDifferentials = nil
  for _, name in ipairs(driveModesDifferentialNames) do
    local c = controller.getController(name)
    if c then
      driveModesDifferentials = driveModesDifferentials or {}
      table.insert(driveModesDifferentials, c)
    end
  end

  driveModesTransfercase = nil
  for _, name in ipairs(driveModesTransfercaseNames) do
    local c = controller.getController(name)
    if c then
      driveModesTransfercase = driveModesTransfercase or {}
      table.insert(driveModesTransfercase, c)
    end
  end

  driveModesTRLT_Range = nil
  for _, name in ipairs(driveModesTRLT_RangeNames) do
    local c = controller.getController(name)
    if c then
      driveModesTRLT_Range = driveModesTRLT_Range or {}
      table.insert(driveModesTRLT_Range, c)
    end
  end
end

M.init = init
M.initLastStage = initLastStage
M.updateGFX = updateGFX
M.toggleDiffs = toggleDiffs
M.toggleTRLT_4WD = toggleTRLT_4WD
M.setTRLT_4WDMode = setTRLT_4WDMode
M.toggleTRLT_Range = toggleTRLT_Range
M.setTRLT_RangeMode = setTRLT_RangeMode
M.serialize = serialize
M.deserialize = deserialize

return M
